@@ -14,6 +14,8 @@
 
 #include "private/gc_priv.h"
 
+#if !defined(PLATFORM_MACH_DEP) && !defined(SN_TARGET_PSP2)
+
 #include <stdio.h>
 
 #ifdef AMIGA
@@ -24,6 +26,89 @@
 # endif
 #endif
 
+#ifdef E2K
+# include <errno.h>
+# include <asm/e2k_syswork.h>
+# include <sys/mman.h>
+# include <sys/syscall.h>
+
+  GC_INNER size_t GC_get_procedure_stack(ptr_t buf, size_t buf_sz) {
+    word new_sz;
+
+    GC_ASSERT(0 == buf_sz || buf != NULL);
+    for (;;) {
+      word stack_ofs;
+
+      new_sz = 0;
+      if (syscall(__NR_access_hw_stacks, E2K_GET_PROCEDURE_STACK_SIZE,
+                  NULL, NULL, 0, &new_sz) == -1) {
+        if (errno != EAGAIN)
+          ABORT_ARG1("Cannot get size of procedure stack",
+                     ": errno= %d", errno);
+        continue;
+      }
+      GC_ASSERT(new_sz > 0 && new_sz % sizeof(word) == 0);
+      if (new_sz > buf_sz)
+        break;
+      /* Immediately read the stack right after checking its size. */
+      stack_ofs = 0;
+      if (syscall(__NR_access_hw_stacks, E2K_READ_PROCEDURE_STACK_EX,
+                  &stack_ofs, buf, new_sz, NULL) != -1)
+        break;
+      if (errno != EAGAIN)
+        ABORT_ARG2("Cannot read procedure stack",
+                   ": new_sz= %lu, errno= %d", (unsigned long)new_sz, errno);
+    }
+    return (size_t)new_sz;
+  }
+
+  ptr_t GC_save_regs_in_stack(void) {
+    __asm__ __volatile__ ("flushr");
+    return NULL;
+  }
+
+  GC_INNER ptr_t GC_mmap_procedure_stack_buf(size_t aligned_sz)
+  {
+    void *buf = mmap(NULL, aligned_sz, PROT_READ | PROT_WRITE,
+                     MAP_PRIVATE | MAP_ANON, 0 /* fd */, 0 /* offset */);
+    if (MAP_FAILED == buf)
+      ABORT_ARG2("Could not map memory for procedure stack",
+                 ": requested %lu bytes, errno= %d",
+                 (unsigned long)aligned_sz, errno);
+    return (ptr_t)buf;
+  }
+
+  GC_INNER void GC_unmap_procedure_stack_buf(ptr_t buf, size_t sz)
+  {
+    if (munmap(buf, ROUNDUP_PAGESIZE(sz)) == -1)
+      ABORT_ARG1("munmap failed (for procedure stack space)",
+                 ": errno= %d", errno);
+  }
+
+# ifdef THREADS
+    GC_INNER size_t GC_alloc_and_get_procedure_stack(ptr_t *pbuf)
+    {
+      /* TODO: support saving from non-zero ofs in stack */
+      ptr_t buf = NULL;
+      size_t new_sz, buf_sz;
+
+      GC_ASSERT(I_HOLD_LOCK());
+      for (buf_sz = 0; ; buf_sz = new_sz) {
+        new_sz = GC_get_procedure_stack(buf, buf_sz);
+        if (new_sz <= buf_sz) break;
+
+        if (EXPECT(buf != NULL, FALSE))
+          GC_INTERNAL_FREE(buf);
+        buf = (ptr_t)GC_INTERNAL_MALLOC_IGNORE_OFF_PAGE(new_sz, PTRFREE);
+        if (NULL == buf)
+          ABORT("Could not allocate memory for procedure stack");
+      }
+      *pbuf = buf;
+      return new_sz;
+    }
+# endif /* THREADS */
+#endif /* E2K */
+
 #if defined(MACOS) && defined(__MWERKS__)
 
 #if defined(POWERPC)
@@ -226,19 +311,22 @@ GC_INNER void GC_with_callee_saves_pushed(void (*fn)(ptr_t, void *),
                                           volatile ptr_t arg)
 {
   volatile int dummy;
-  void * volatile context = 0;
+  volatile ptr_t context = 0;
 
 # if defined(HAVE_PUSH_REGS)
     GC_push_regs();
+# elif defined(EMSCRIPTEN)
+    /* No-op, "registers" are pushed in GC_push_other_roots().  */
 # else
 #   if defined(UNIX_LIKE) && !defined(NO_GETCONTEXT)
       /* Older versions of Darwin seem to lack getcontext().    */
       /* ARM and MIPS Linux often doesn't support a real        */
       /* getcontext().                                          */
+      static signed char getcontext_works = 0; /* (-1) - broken, 1 - works */
       ucontext_t ctxt;
 #     ifdef GETCONTEXT_FPU_EXCMASK_BUG
         /* Workaround a bug (clearing the FPU exception mask) in        */
-        /* getcontext on Linux/x86_64.                                  */
+        /* getcontext on Linux/x64.                                     */
 #       ifdef X86_64
           /* We manipulate FPU control word here just not to force the  */
           /* client application to use -lm linker option.               */
@@ -253,12 +341,17 @@ GC_INNER void GC_with_callee_saves_pushed(void (*fn)(ptr_t, void *),
 #       endif
 #     endif
 
-      if (getcontext(&ctxt) < 0) {
-        WARN("getcontext failed:"
-             " using another register retrieval method...\n", 0);
-        /* E.g., to workaround a bug in Docker ubuntu_32bit.    */
-      } else {
-        context = &ctxt;
+      if (getcontext_works >= 0) {
+        if (getcontext(&ctxt) < 0) {
+          WARN("getcontext failed:"
+               " using another register retrieval method...\n", 0);
+          /* getcontext() is broken, do not try again.          */
+          /* E.g., to workaround a bug in Docker ubuntu_32bit.  */
+        } else {
+          context = (ptr_t)&ctxt;
+        }
+        if (EXPECT(0 == getcontext_works, FALSE))
+          getcontext_works = context != NULL ? 1 : -1;
       }
 #     ifdef GETCONTEXT_FPU_EXCMASK_BUG
 #       ifdef X86_64
@@ -276,11 +369,14 @@ GC_INNER void GC_with_callee_saves_pushed(void (*fn)(ptr_t, void *),
             ABORT("feenableexcept failed");
 #       endif
 #     endif /* GETCONTEXT_FPU_EXCMASK_BUG */
-#     if defined(SPARC) || defined(IA64)
+#     if defined(E2K) || defined(IA64) || defined(SPARC)
         /* On a register window machine, we need to save register       */
         /* contents on the stack for this to work.  This may already be */
         /* subsumed by the getcontext() call.                           */
-        GC_save_regs_ret_val = GC_save_regs_in_stack();
+#       if defined(IA64) || defined(SPARC)
+          GC_save_regs_ret_val =
+#       endif
+            GC_save_regs_in_stack();
 #     endif
       if (NULL == context) /* getcontext failed */
 #   endif /* !NO_GETCONTEXT */
@@ -290,14 +386,17 @@ GC_INNER void GC_with_callee_saves_pushed(void (*fn)(ptr_t, void *),
         /* force callee-save registers and register windows onto        */
         /* the stack.                                                   */
         __builtin_unwind_init();
+#     elif defined(NO_CRT) && defined(MSWIN32)
+        CONTEXT ctx;
+        RtlCaptureContext(&ctx);
 #     else
         /* Generic code                          */
         /* The idea is due to Parag Patel at HP. */
         /* We're not sure whether he would like  */
         /* to be acknowledged for it or not.     */
         jmp_buf regs;
-        register word * i = (word *) &regs;
-        register ptr_t lim = (ptr_t)(&regs) + (sizeof regs);
+        word *i = (word *)&regs[0];
+        ptr_t lim = (ptr_t)(&regs[0]) + sizeof(regs);
 
         /* Setjmp doesn't always clear all of the buffer.               */
         /* That tends to preserve garbage.  Clear it.                   */
@@ -306,7 +405,8 @@ GC_INNER void GC_with_callee_saves_pushed(void (*fn)(ptr_t, void *),
         }
 #       if defined(MSWIN32) || defined(MSWINCE) || defined(UTS4) \
            || defined(OS2) || defined(CX_UX) || defined(__CC_ARM) \
-           || defined(LINUX) || defined(EWS4800) || defined(RTEMS)
+           || defined(LINUX) || defined(EWS4800) || defined(RTEMS) \
+           || defined(DJGPP)
           (void) setjmp(regs);
 #       else
           (void) _setjmp(regs);
@@ -317,21 +417,13 @@ GC_INNER void GC_with_callee_saves_pushed(void (*fn)(ptr_t, void *),
 #     endif /* !HAVE_BUILTIN_UNWIND_INIT */
     }
 # endif /* !HAVE_PUSH_REGS */
-  /* FIXME: context here is sometimes just zero.  At the moment the     */
+  /* TODO: context here is sometimes just zero.  At the moment, the     */
   /* callees don't really need it.                                      */
-  fn(arg, context);
+  fn(arg, (/* no volatile */ void *)context);
   /* Strongly discourage the compiler from treating the above   */
   /* as a tail-call, since that would pop the register          */
   /* contents before we get a chance to look at them.           */
-  GC_noop1((word)(&dummy));
+  GC_noop1(COVERT_DATAFLOW(&dummy));
 }
 
-#if defined(ASM_CLEAR_CODE)
-# ifdef LINT
-    ptr_t GC_clear_stack_inner(ptr_t arg, word limit)
-    {
-      return limit ? arg : 0; /* use both arguments */
-    }
-    /* The real version is in a .S file */
-# endif
-#endif /* ASM_CLEAR_CODE */
+#endif /* !PLATFORM_MACH_DEP && !SN_TARGET_PSP2 */
