@@ -136,14 +136,14 @@ STATIC void GC_extend_size_map(size_t i)
   /* For these larger sizes, we use an even number of granules.         */
   /* This makes it easier to, e.g., construct a 16-byte-aligned         */
   /* allocator even if GRANULE_BYTES is 8.                              */
-  granule_sz = (granule_sz + 1) & ~1;
+  granule_sz = (granule_sz + 1) & ~(size_t)1;
   if (granule_sz > MAXOBJGRANULES)
     granule_sz = MAXOBJGRANULES;
 
   /* If we can fit the same number of larger objects in a block, do so. */
   number_of_objs = HBLK_GRANULES / granule_sz;
   GC_ASSERT(number_of_objs != 0);
-  granule_sz = (HBLK_GRANULES / number_of_objs) & ~1;
+  granule_sz = (HBLK_GRANULES / number_of_objs) & ~(size_t)1;
 
   byte_sz = GRANULES_TO_BYTES(granule_sz) - EXTRA_BYTES;
                         /* We may need one extra byte; do not always    */
@@ -154,7 +154,7 @@ STATIC void GC_extend_size_map(size_t i)
 }
 
 /* Allocate lb bytes for an object of kind k.           */
-/* Should not be used to directly to allocate objects   */
+/* Should not be used to directly allocate objects      */
 /* that require special handling on allocation.         */
 GC_INNER void * GC_generic_malloc_inner(size_t lb, int k)
 {
@@ -199,9 +199,11 @@ GC_INNER void * GC_generic_malloc_inner(size_t lb, int k)
         obj_link(op) = 0;
         GC_bytes_allocd += GRANULES_TO_BYTES((word)lg);
     } else {
-        op = (ptr_t)GC_alloc_large_and_clear(ADD_SLOP(lb), k, 0);
+        size_t lb_adjusted = ADD_SLOP(lb);
+
+        op = (ptr_t)GC_alloc_large_and_clear(lb_adjusted, k, 0 /* flags */);
         if (op != NULL)
-            GC_bytes_allocd += lb;
+            GC_bytes_allocd += lb_adjusted;
     }
 
     return op;
@@ -210,10 +212,10 @@ GC_INNER void * GC_generic_malloc_inner(size_t lb, int k)
 #if defined(DBG_HDRS_ALL) || defined(GC_GCJ_SUPPORT) \
     || !defined(GC_NO_FINALIZATION)
   /* Allocate a composite object of size n bytes.  The caller           */
-  /* guarantees that pointers past the first page are not relevant.     */
+  /* guarantees that pointers past the first hblk are not relevant.     */
   GC_INNER void * GC_generic_malloc_inner_ignore_off_page(size_t lb, int k)
   {
-    word lb_adjusted;
+    size_t lb_adjusted;
     void * op;
 
     GC_ASSERT(I_HOLD_LOCK());
@@ -265,7 +267,11 @@ GC_API GC_ATTR_MALLOC void * GC_CALL GC_generic_malloc(size_t lb, int k)
         LOCK();
         result = (ptr_t)GC_alloc_large(lb_rounded, k, 0);
         if (0 != result) {
-          if (GC_debugging_started) {
+          if (GC_debugging_started
+#             ifndef THREADS
+                || init
+#             endif
+             ) {
             BZERO(result, n_blocks * HBLKSIZE);
           } else {
 #           ifdef THREADS
@@ -280,9 +286,13 @@ GC_API GC_ATTR_MALLOC void * GC_CALL GC_generic_malloc(size_t lb, int k)
           GC_bytes_allocd += lb_rounded;
         }
         UNLOCK();
-        if (init && !GC_debugging_started && 0 != result) {
-            BZERO(result, n_blocks * HBLKSIZE);
-        }
+#       ifdef THREADS
+          if (init && !GC_debugging_started && result != NULL) {
+            /* Clear the rest (i.e. excluding the initial 2 words). */
+            BZERO((word *)result + 2,
+                  n_blocks * HBLKSIZE - 2 * sizeof(word));
+          }
+#       endif
     }
     if (0 == result) {
         return((*GC_get_oom_fn())(lb));
@@ -460,26 +470,46 @@ GC_API GC_ATTR_MALLOC void * GC_CALL GC_malloc_uncollectable(size_t lb)
     STATIC ptr_t GC_libpthread_end = 0;
     STATIC ptr_t GC_libld_start = 0;
     STATIC ptr_t GC_libld_end = 0;
+    static GC_bool lib_bounds_set = FALSE;
 
-    STATIC void GC_init_lib_bounds(void)
+    GC_INNER void GC_init_lib_bounds(void)
     {
       IF_CANCEL(int cancel_state;)
+      DCL_LOCK_STATE;
+
+      /* This test does not need to ensure memory visibility, since     */
+      /* the bounds will be set when/if we create another thread.       */
+      if (EXPECT(lib_bounds_set, TRUE)) return;
 
-      if (GC_libpthread_start != 0) return;
       DISABLE_CANCEL(cancel_state);
       GC_init(); /* if not called yet */
+#     if defined(GC_ASSERTIONS) && defined(GC_ALWAYS_MULTITHREADED)
+        LOCK(); /* just to set GC_lock_holder */
+#     endif
       if (!GC_text_mapping("libpthread-",
                            &GC_libpthread_start, &GC_libpthread_end)) {
+        /* Some libc implementations like bionic, musl and glibc 2.34   */
+        /* do not have libpthread.so because the pthreads-related code  */
+        /* is located in libc.so, thus potential calloc calls from such */
+        /* code are forwarded to real (libc) calloc without any special */
+        /* handling on the libgc side.  Checking glibc version at       */
+        /* compile time to turn off the warning seems to be fine.       */
+        /* TODO: Remove GC_text_mapping() call for this case.           */
+#       if defined(__GLIBC__) \
+           && (__GLIBC__ < 2 || (__GLIBC__ == 2 && __GLIBC_MINOR__ < 34))
           WARN("Failed to find libpthread.so text mapping: Expect crash\n", 0);
           /* This might still work with some versions of libpthread,      */
-          /* so we don't abort.  Perhaps we should.                       */
-          /* Generate message only once:                                  */
-            GC_libpthread_start = (ptr_t)1;
+          /* so we do not abort.                                          */
+#       endif
       }
       if (!GC_text_mapping("ld-", &GC_libld_start, &GC_libld_end)) {
           WARN("Failed to find ld.so text mapping: Expect crash\n", 0);
       }
+#     if defined(GC_ASSERTIONS) && defined(GC_ALWAYS_MULTITHREADED)
+        UNLOCK();
+#     endif
       RESTORE_CANCEL(cancel_state);
+      lib_bounds_set = TRUE;
     }
 # endif /* GC_LINUX_THREADS */
 
@@ -492,14 +522,9 @@ GC_API GC_ATTR_MALLOC void * GC_CALL GC_malloc_uncollectable(size_t lb)
       /* libpthread allocated some memory that is only pointed to by    */
       /* mmapped thread stacks.  Make sure it is not collectible.       */
       {
-        static GC_bool lib_bounds_set = FALSE;
         ptr_t caller = (ptr_t)__builtin_return_address(0);
-        /* This test does not need to ensure memory visibility, since   */
-        /* the bounds will be set when/if we create another thread.     */
-        if (!EXPECT(lib_bounds_set, TRUE)) {
-          GC_init_lib_bounds();
-          lib_bounds_set = TRUE;
-        }
+
+        GC_init_lib_bounds();
         if (((word)caller >= (word)GC_libpthread_start
              && (word)caller < (word)GC_libpthread_end)
             || ((word)caller >= (word)GC_libld_start
@@ -600,7 +625,7 @@ GC_API void GC_CALL GC_free(void * p)
         LOCK();
         GC_bytes_freed += sz;
         if (IS_UNCOLLECTABLE(knd)) GC_non_gc_bytes -= sz;
-                /* Its unnecessary to clear the mark bit.  If the       */
+                /* It's unnecessary to clear the mark bit.  If the      */
                 /* object is reallocated, it doesn't matter.  O.w. the  */
                 /* collector will do it, since it's on a free list.     */
         if (ok -> ok_init && EXPECT(sz > sizeof(word), TRUE)) {
@@ -678,7 +703,7 @@ GC_API void GC_CALL GC_free(void * p)
 #   define REDIRECT_FREE_F REDIRECT_FREE
 # endif
 
-  void free(void * p)
+  void free(void * p GC_ATTR_UNUSED)
   {
 #   ifndef IGNORE_FREE
 #     if defined(GC_LINUX_THREADS) && !defined(USE_PROC_FOR_LIBRARIES)
